Item = {}
Container = Item

local Item = Item
local meta = {__index = Item}

function Item.IsItem(object)
	return getmetatable(object) == meta
end

function meta:__tostring()
	return "Item ["..self.UID.."]["..self:GetDataName().." "..self:GetAmount().."]"
end

function meta:__eq(other)
	return other ~= nil and Item.IsItem(other) and other.UID == self.UID
end

-- This is for the NDB Serialize function. It's not a real metamethod.
function Item:_serialize()
	return "Item("..Serialize(self, true)..")"
end
Item.Serialize = Item._serialize

-- NonStrict functions here are just for convenience.
function Item:AddItemNonStrict(object, amount, autostack)
	if not object or not self:IsContainer() then return false end

	if Item.IsItem(object) then
		return self:AddItem(object, autostack)
	end

	local objecttype = type(object)

	if objecttype == "string" then
		local item = Item(nil, object, amount)
		if item then
			return self:AddItem(item, autostack)
		end
	elseif type(object) == "Entity" and not object:IsRemoving() and object:GetItem() and self:AddItem(object:GetItem(), autostack) then
		object:Remove()
		return true
	end

	return false
end
Item.GiveItemNonStrict = Item.AddItemNonStrict

function Item:RemoveItemNonStrict(object, amount)
	if not object or not self:IsContainer() then return false end

	if Item.IsItem(object) then
		return self:RemoveItem(object, amount)
	end

	amount = amount or 1

	if type(object) == "string" then
		for _, item in pairs(self:GetChildren()) do
			if item:GetDataName() == object and item:GetAmount() >= amount then
				return self:RemoveItem(item, amount)
			end
		end
	end

	return false
end
Item.TakeItemNonStrict = Item.RemoveItemNonStrict
Item.DestroyItemNonStrict = Item.RemoveItemNonStrict

function Item:HasItemNonStrict(object, amount)
	if amount then
		return self:GetItemAmountNonStrict(object) >= amount
	end

	if not object or not self:IsContainer() then return false end

	if Item.IsItem(object) then
		return self:HasItem(object)
	end

	if type(object) == "string" then
		for _, child in pairs(self:GetChildren(true)) do
			if child:GetDataName() == object or child:IsContainer() and child:HasItemNonStrict(object) then return true end
		end
	end

	return false
end

function Item:GetItemAmountNonStrict(object, inclusive)
	local amount = 0
	if not object or not self:IsContainer() then return amount end

	for _, child in pairs(self:GetChildren(inclusive)) do
		if child:GetDataName() == object then
			amount = amount + child:GetAmount()
		end
	end

	return amount
end
Item.GetItemCountNonStrict = Item.GetItemAmountNonStrict
Item.ItemAmountNonStrict = Item.GetItemAmountNonStrict
Item.ItemCountNonStrict = Item.GetItemAmountNonStrict

-- Returns the total amount of items in this container. inclusive means don't search children.
function Item:GetTotalItemCount(inclusive)
	local amount = 0
	if not self:IsContainer() then return amount end

	for _, child in pairs(self:GetChildren(inclusive)) do
		amount = amount + child:GetAmount()
	end

	return amount
end
Item.TotalItemCount = Item.GetTotalItemCount

-- Returns the total amount of item objects in this container. inclusive means don't search children.
function Item:GetTotalItemObjectCount(inclusive)
	local amount = 0
	if not self:IsContainer() then return amount end

	for _, child in pairs(self:GetChildren(inclusive)) do
		amount = amount + 1
	end

	return amount
end
Item.TotalItemObjectCount = Item.GetTotalItemObjectCount
Item.GetItemObjectCount = Item.GetTotalItemObjectCount

-- Attempt to combine with like items in this container. Returns true on success.
function Item:StackWithOthersInContainer(parent)
	parent = parent or self:GetParent()
	if not parent then return false end

	for _, child in pairs(parent:GetChildren(true)) do
		if child:IsStackable() and child:CanStack(self) then
			self:SetAmount(self:GetAmount() + child:GetAmount())
			child:Destroy()
			return true
		end
	end
end

-- Returns the number of item objects we can hold. nil is infinite, which is a bad idea. Remember, this is an OBJECT count.
function Item:GetCapacity()
	local itemdata = self:GetItemData()
	return itemdata.GetCapacity and itemdata.GetCapacity(self) or self.ItemCapacity
end

-- Returns the maximum amount of mass we can hold. nil is infinite.
function Item:GetMassCapacity()
	local itemdata = self:GetItemData()
	return itemdata.GetMassCapacity and itemdata.GetMassCapacity(self) or self.MassCapacity
end

-- Adds an item to us.
-- Returns true on success.
function Item:AddItem(other, autostack, nosync)
	if self:IsContainer() then
		local capacity = self:GetCapacity()
		if capacity and capacity ~= -1 and capacity <= self:TotalItemObjectCount() then
			return false
		end

		-- TODO: This is bugged.
		--[[local masscapacity = self:GetMassCapacity()
		if masscapacity and masscapacity ~= -1 and masscapacity < (self:GetMass() - self:GetMass(true)) + other:GetMass() then
			return false
		end]]

		other:SetParent(self)

		if autostack then
			other:StackWithOthersInContainer()
		end

		if not nosync then
			self:RadiusSync()
		end

		return true
	--else
		-- TODO: Try stacking
	end

	return false
end

function Item:CanStack(other)
	if other == self then return false end

	local itemdata = self:GetItemData()
	local ret = itemdata.CanStack and itemdata.CanStack(self, other)
	if ret ~= nil then
		return ret
	end

	return self:GetDataName() == other:GetDataName() and self:GetName() == other:GetName() and self:GetModel() == other:GetModel() and (self.MaxStack == -1 or self.MaxStack >= self:GetAmount() + other:GetAmount())
end

function Item:IsStackable()
	return self.MaxStack == -1 or self.MaxStack >= 2
end

-- If we contain Item other, removes it from this item.
-- Make sure you do something with the item you remove or it may be garbage collected.
-- You can also pass amount as the second argument.
-- Returns true on success.
function Item:RemoveItem(other, amount)
	if other:IsInside(self) then
		if amount and amount < other:GetAmount() then
			other:SetAmount(other:GetAmount() - amount)
		else
			other:SetParent(nil)
		end

		self:RadiusSync()

		return true
	end

	return false
end

-- Is this item inside item other? Checks multiple levels of containers.
function Item:IsInside(other)
	local parent = self:GetParent()
	if parent then
		if parent == other then
			return true
		else
			return parent:IsInside(other)
		end
	end

	return false
end

-- Our dataname. Same as the item's file name in the items folder.
function Item:GetDataName()
	return self.DataName or "rock"
end

-- Usable? This function should exist both server-side and client-side, even if the function is empty.
function Item:GetUsable()
	return self:GetItemData().OnUse ~= nil
end
Item.IsUsable = Item.GetUsable

-- Change our name. If this is nil then it will use either the item file's name or the dataname.
function Item:SetName(name)
	self.Name = name
end

-- Get our name. Doesn't take in to consideration item amount or grammar.
function Item:GetName()
	return self.Name or self:GetItemData().Name or self:GetDataName()
end

-- A nice looking name. Instead of "rock" it would display "a rock" or "4 rocks" if GetAmount was 4.
function Item:GetDisplayName()
	return util.NameByAmount(self:GetName(), self:GetAmount())
end

-- Amount in this item.
function Item:GetAmount()
	return self.Amount or 1
end

-- Sets the amount. Does not check against MaxStack. That should be done on a case-by-case basis.
function Item:SetAmount(amount)
	assert(type(amount) == "number", "SetAmount expects a number.")

	if self.Amount ~= amount then
		self.Amount = amount

		if amount <= 0 then
			self:Destroy()
		else
			self:RadiusSync()
		end
	end
end

function Item:SetModel(mdl)
	if self.Model ~= mdl then
		self.Model = mdl
		self:RadiusSync()
	end
end

function Item:GetModel()
	return self.Model
end

-- Gets every item under us.
function Item:GetChildren(thisonly)
	local tab = {}

	if self:IsContainer() then
		for _, item in pairs(self:GetContainer()) do
			tab[#tab + 1] = item
			if not thisonly and item:IsContainer() then
				table.Add(tab, item:GetChildren())
			end
		end
	end

	return tab
end

-- If this item is a container, does it hold Item other? Searches any children containers as well.
function Item:ContainsItem(other)
	if self:IsContainer() then
		for _, item in pairs(self:GetContainer()) do
			if other == item or item:ContainsItem(other) then
				return true
			end
		end
	end

	return false
end

-- Is this item a container?
function Item:IsContainer()
	return self.Container ~= nil
end

-- Returns our container if we have one.
function Item:GetContainer()
	return self.Container
end

-- This isn't really used anywhere else.
function Item:GetItem(object)
	if self:ContainsItem(object) then return object end
end

-- Search for an item by Item or DataName.
function Item:GetItemNonStrict(object, bInclusive)
	if not self:IsContainer() then return end

	if Item.IsItem(object) then
		return self:GetItem(object)
	end

	if type(object) == "string" then -- Asking for a DataName. Just return the first entry we get (if any).
		for _, item in pairs(self:GetContainer()) do
			if item:GetDataName() == object then
				return item
			end
		end

		if not bInclusive then
			-- Not found in this level, search children.
			for _, item in pairs(self:GetContainer()) do
				local ret = item:GetItem(object)
				if ret then return ret end
			end
		end
	elseif type(object) == "number" then -- Asking for a UID. Return the global item only if we contain it.
		return self:GetItem(Items[object])
	end
end

-- Gets the entity that contains us. It can be that we're the entity's item or we're contained in a container the entity owns.
function Item:GetRootEntity()
	local root = self:GetRootOrSelf()
	for _, ent in pairs(ents.GetAll()) do
		if ent:GetItem() == root then
			return ent
		end
	end

	return NULL
end
Item.GetRootOwner = Item.GetRootEntity

-- Items don't hold pointers to their entity. You usually don't need to call this.
function Item:GetEntity()
	for _, ent in pairs(ents.GetAll()) do
		if ent:IsValid() and ent:GetContainer() == self then
			return ent
		end
	end

	return NULL
end
Item.GetOwner = Item.GetEntity

function Item:SetDroppable(droppable)
	if droppable == false then
		self.Undroppable = true
	else
		self.Undroppable = nil
	end
	self:RadiusSync()
end

function Item:GetDroppable()
	if self.Undroppable then return false end

	local itemdata = self:GetItemData()
	if itemdata.IsDroppable then
		local ret = itemdata.IsDroppable(item)
		if ret ~= nil then return ret end
	end

	return true
end
Item.IsDroppable = Item.GetDroppable

-- Move the item around in a container. Nothing to do with entities.
function Item:Move(x, y)
	self.X = math.Clamp(math.ceil(x), 0, 2048)
	self.Y = math.Clamp(math.ceil(y), 0, 2048)

	self:RadiusSync()
end

function Item:InRange(ent)
	local entpos = ent:IsPlayer() and ent:NewEyePos() or ent:EyePos()
	local owner = self:GetRootEntity()
	if owner == ent then return true end

	return owner:IsValid() and owner:NearestPoint(entpos):Distance(entpos) < ITEM_USABLEDISTANCE
end

-- TODO: This needs to be better.
function Item:InView(ent)
	local entpos = ent:IsPlayer() and ent:NewEyePos() or ent:EyePos()
	local owner = self:GetRootEntity()
	if owner == ent then return true end

	return owner:IsValid() and not util.TraceLine({start = entpos, endpos = owner:NearestPoint(entpos), mask = MASK_SOLID_BRUSHONLY}).Hit
end

function Item:InRangeAndView(ent)
	local entpos = ent:IsPlayer() and ent:NewEyePos() or ent:EyePos()
	local owner = self:GetRootEntity()
	if owner == ent then return true end

	if owner:IsValid() then
		local nearest = owner:NearestPoint(entpos)
		return nearest:Distance(entpos) < ITEM_USABLEDISTANCE
		and not util.TraceLine({start = entpos, endpos = nearest, mask = MASK_SOLID_BRUSHONLY}).Hit
	end

	return false
end

-- Can this player interact with us?
function Item:IsUsableBy(pl)
	if not pl:Alive() or not self:InRangeAndView(pl) then return false end

	local owner = self:GetRootEntity()

	if self.WearableSlot then
		return owner:IsValid() and (owner == pl or owner:GetOwner() == pl)
	end

	return not owner:IsValid() or not owner:IsPlayer() or owner == pl
end

-- Can this player drop us from our current container?
function Item:IsDroppableBy(pl)
	return self:IsMoveableBy(pl)
	and not self.Bound and not self.Cursed
end

-- Can this player view us?
function Item:IsViewableBy(pl)
	return self:InRangeAndView(pl)
end

-- Can this player move us within the current container?
function Item:IsMoveableBy(pl)
	if self.LockedDown or self.Moveable == false or not pl:Alive() or not self:InRangeAndView(pl) then return false end

	local owner = self:GetRootEntity()

	if owner.LockedDown then return false end -- Can't move an item that has their entity locked down.

	return not owner:IsValid() or -- Item in no container??
	not owner:IsPlayer() or -- Entity in the open.
	owner == pl -- In our player inventory.
	and not self.Hidden -- Can't move hidden items.
end

-- Can this player transfer us to the new container?
function Item:IsTransferableBy(pl, newcontainer)
	return newcontainer ~= nil
	and self.UID ~= newcontainer.UID -- Can't transfer ourself to ourself.
	and not self:ContainsItem(newcontainer) -- Can't transfer ourself to a container we contain.
	and self:IsMoveableBy(pl) and newcontainer:IsUsableBy(pl) -- Can't transfer something we can't move.
	and not self.Bound and not self.Cursed -- Can't transfer items with these special flags.
end

-- Creates an entity and sets this Item object to that entity.
-- Ironically, this doesn't actually call :Spawn() on the entity. That must be done yourself.
function Item:SpawnEntity()
	local curent = self:GetEntity()
	if curent:IsValid() then
		return curent
	end

	local class = self:GetEntityClass()
	if scripted_ents.GetStored(class) then
		local ent = ents.Create(class)
		if ent:IsValid() then
			self:SetParent(nil) -- Entities can never be in a container.
			ent:SetItem(self)
			return ent
		end
	end

	return NULL
end
Item.Spawn = Item.SpawnEntity
Item.SpawnItem = Item.SpawnEntity

-- Sets our color then forwards it to our entity.
function Item:SetColor(col)
	self.Color = col

	local ent = self:GetEntity()
	if ent:IsValid() then ent:SetColor(col.r, col.g, col.b, col.a) end

	self:Sync()
end

-- Sets our material then forwards it to our entity.
function Item:SetMaterial(mat)
	self.Material = mat

	local ent = self:GetEntity()
	if ent:IsValid() then ent:SetMaterial(mat) end

	self:Sync()
end

-- Sets our mass then forwards it to our entity.
function Item:SetMass(mass)
	self.Mass = mass

	local ent = self:GetEntity()
	if ent:IsValid() then
		local phys = ent:GetPhysicsObject()
		if phys:IsValid() then
			phys:SetMass(math.max(1, mass * self:GetAmount()))
		end
	end

	self:Sync()
end

function SpawnItem(class, amount)
	return Item(nil, class, amount):Spawn()
end

-- Returns a copy of this item. The only thing changed is the UID.
function Item:Copy()
	local data = table.Copy(self)
	data.UID = nil
	return Item(data)
end

local function DoOnContentsChanged(func, container, item)
	if container and container:IsItem() then
		func(container, item)
	end
end
function Item:CallParentContentsChanged()
	local oldcontainer = self:GetRoot()
	if oldcontainer and oldcontainer:IsItem() and oldcontainer:IsContainer() then
		local itemdata = oldcontainer:GetItemData()
		if itemdata and itemdata.OnContentsChanged then
			timer.Simple(0, DoOnContentsChanged, itemdata.OnContentsChanged, oldcontainer, self)
		end
	end
end

-- Gets the biggest container that ultimately contains us.
function Item:GetRoot()
	local parent = self:GetParent()
	if parent then
		return parent:GetRoot() or parent
	end

	--[[local parent = self:GetParent()
	if parent then
		local ret = parent:GetParent()
		if ret then return ret end

		return parent
	end]]
end

function Item:GetRootOrSelf()
	return self:GetRoot() or self
end

if SERVER then
-- Sync this item with the owner, if they're a player.
-- Optionally allows you to specify a player instead of sending to the owner.
function Item:Synchronize(pl)
	pl = pl or self:GetRootEntity()
	if pl:IsValid() and pl:IsPlayer() then
		pl:SendLongString(LONGSTRING_UPDATEITEM, self:Serialize())
	end
end
Item.Synch = Item.Synchronize
Item.Syncronize = Item.Synchronize
Item.Sync = Item.Syncronize

-- Syncs with every single client on the server. Probably shouldn't be used that much.
function Item:GlobalSync()
	local serialized = self:Serialize()
	for _, pl in pairs(player.GetAll()) do
		pl:SendLongString(LONGSTRING_UPDATEITEM, serialized)
	end
end

-- This is mostly used for syncing to anyone who can POTENTIALLY be looking at us. It's a bad workaround until a system is made to decide who is looking at an item at any given time.
function Item:RadiusSync(radius)
	if self:IsContainer() or self:GetParent() then -- Don't send unless we're in a container or we are a container.
		local rootentity = self:GetRootEntity()
		if rootentity:IsValid() then
			local serialized = self:Serialize()
			if rootentity:IsPlayer() then -- In a player inventory, only sync to that person.
				rootentity:SendLongString(LONGSTRING_UPDATEITEM, serialized)
			else
				for _, pl in pairs(ents.FindInSphere(rootentity:GetPos(), radius or 1024)) do
					if pl:IsPlayer() then
						pl:SendLongString(LONGSTRING_UPDATEITEM, serialized)
					end
				end
			end
		end
	end
end
end

if CLIENT then
function Item:Synchronize(pl)
end
Item.Synch = Item.Synchronize
Item.Syncronize = Item.Synchronize
Item.Sync = Item.Syncronize
Item.GlobalSync = Item.Synchronize
Item.RadiusSync = Item.Synchronize
end

-- Utility to get our entity class, if we had one.
function Item:GetEntityClass()
	return "item_"..self:GetDataName()
end

-- Returns our itemdata. This is what you get from the scripts in the items folder.
function Item:GetItemData()
	return GetItemData(self.DataName)
end

-- Destroys this item. If it's a container, destroys any children.
-- If we have an entity, the entity is removed.
function Item:Destroy()
	self:SetParent(nil)
	for _, item in pairs(self:GetChildren(true)) do
		item:Destroy()
	end

	self:CallParentContentsChanged()

	self:RemoveEntity()
end
Item.Remove = Item.Destroy

-- Returns the data of our baseclass if we have one.
function Item:GetBaseClass()
	return GetItemData(self.Base)
end

function Item:RemoveEntity()
	self:CallParentContentsChanged()

	local ent = self:GetEntity()
	if ent:IsValid() and not ent:IsCharacter() then
		ent:Remove()
	end
end

function Item:RemoveAllEntities()
	self:RemoveEntity()
	for _, child in pairs(self:GetChildren()) do
		child:RemoveEntity()
	end
end

function Item:OnRootEntityChanged()
	self:RemoveAllEntities()
	self:RadiusSync()
end

-- Sets our parent Item to parent.
-- parent must be a container.
-- This function will also remove us from our current container, if we have one.
-- Removes our entity if we have one.
-- Returns true on success.
function Item:SetParent(parent)
	local curparent = self:GetParent()
	if curparent == parent then return true end -- Skip it all if the new parent is the same as our current.

	if not parent or parent:IsContainer() then -- Either the parent is nothing or a container.
		local oldentity = self:GetRootEntity()

		if parent then
			self.Parent = parent.UID
			self:RemoveEntity()
		else
			self.Parent = nil
		end

		if curparent then
			curparent.Container[self.UID] = nil
			curparent:RadiusSync()
		end

		if parent then
			parent.Container[self.UID] = self
			parent:RadiusSync()
		end

		if curparent ~= parent then
			if curparent then
				local cpitemdata = curparent:GetItemData()
				if cpitemdata and cpitemdata.OnContentsChanged then
					cpitemdata.OnContentsChanged(curparent, self)
				end
			end

			if parent then
				local pitemdata = parent:GetItemData()
				if pitemdata and pitemdata.OnContentsChanged then
					pitemdata.OnContentsChanged(parent, self)
				end
			end

			local itemdata = self:GetItemData()
			if itemdata and itemdata.OnParentChanged then
				itemdata.OnParentChanged(self, parent, curparent)
			end
		end

		if self:GetRootEntity() ~= oldentity then
			self:OnRootEntityChanged()
		end

		return true
	end

	return false
end

-- Returns the Item object that holds us, if any.
function Item:GetParent()
	if self.Parent then
		return Items[self.Parent]
	end
end

-- Returns the TOTAL mass of this item.
-- This includes the mass of any children if we're a container as well as our own mass.
function Item:GetMass(onlyself)
	if onlyself or not self:IsContainer() then return self.Mass end

	local mass = self.Mass

	for _, child in pairs(self:GetChildren()) do
		mass = mass + child:GetMass()
	end

	return mass
end

function Item:IsStaff()
	local swep = weapons.GetStored(self:GetDataName())
	return swep and swep.Base and string.sub(swep.Base, 1, 18) == "weapon__base_staff"
end
Item.IsMagic = Item.IsStaff
Item.IsWand = Item.IsStaff

function Item:IsBow()
	local swep = weapons.GetStored(self:GetDataName())
	return swep and swep.Base and string.sub(swep.Base, 1, 16) == "weapon__base_bow"
end

function Item:IsMelee()
	local swep = weapons.GetStored(self:GetDataName())
	return swep and swep.Base and string.sub(swep.Base, 1, 18) == "weapon__base_melee"
end

function Item:IsRanged()
	return self:IsBow() or self:IsStaff()
end

if CLIENT then
	-- Returns an item panel for this item.
	function Item:GetItemPanel(parent)
		return ItemPanel(self, parent)
	end
	Item.ItemPanel = Item.GetItemPanel

	NDB.AddContentsCallback(LONGSTRING_UPDATEITEM, function(contents)
		VERBOSE("updateitem", contents)
		RunString("gamemode.Call(\"ItemReceived\", "..contents..")")
	end)
end

-- Use Item([data], [dataname], [amount]) instead of this Item.new. Handled by the __call metamethod below.
-- All 3 arguments are optional.
-- Returns an Item object on success, otherwise returns nil.
-- If an Item already exists with data's UID then it returns the existing Item object (can't be more than one Item object using the same UID).
function Item:new(data, dataname, amount)
	local item

	local bIsNew = false
	if data and data.UID and Items[data.UID] ~= nil then
		item = Items[data.UID]
		for k, v in pairs(item) do
			item[k] = nil
		end
		--[[for k, v in pairs(item) do
			if data[k] == nil then
				item[k] = nil
			end
		end]]
	else
		item = {}
		setmetatable(item, meta)
	end

	if data then
		for k, v in pairs(data) do
			item[k] = v
		end
	end

	bIsNew = item.Created == nil

	item.UID = item.UID or GetUID()
	item.DataName = item.DataName or dataname or "rock"

	if not GetItemData(item.DataName) then return end
	
	if bIsNew then
		item.Created = os.time()

		item.X = math.random(0, 400)
		item.Y = math.random(0, 400)

		local copydata = GetItemData(dataname, true)
		if copydata then
			for k, v in pairs(copydata) do
				item[k] = v
			end
		end

		local itemdata = GetItemData(item.DataName)
		if itemdata then
			if 0 < itemdata.MaxStack then
				item:SetAmount(math.min(itemdata.MaxStack, item:GetAmount() or amount))
			else
				item:SetAmount(item.Amount or amount or 1)
			end

			if itemdata.OnCreated then
				itemdata.OnCreated(item)
			end
		end
	end

	Items[item.UID] = item

	return item
end

setmetatable(Item, {__call = Item.new})
